##############################################################################
#
# This is the history of Michele Cappellari modifications
# to Daniel Foreman-Mackey corner_plot routine.
#
# V1.0.0: Included "like" and "xstry" optional inputs
#   to show individual points coloured by their likelihood
#   and to show all points tried by MCMC.
#   Michele Cappellari, Oxford, 12 January 2014
# V1.1.0: Added "init" keyword to simply initialize an empty window.
#   MC, Oxford, 10 October 2014
# V1.1.1: Check that input sizes of "extents" and "labels" match.
#   MC, Portsmouth, 15 October 2014
# V1.1.2: Only show unique dots in "xs" and "like", but properly include
#   duplicates in histograms. MC, Oxford, 4 November 2014
# V1.1.3: Allow for scaling of axes labels. 
#   Fix program stop when values are the same within numerical accuracy
#   but np.std(x) is not exactly zero. MC, Oxford, 10 December 2014
# V1.1.4: Fix potential program stop with few input values.
#   MC, Oxford, 20 November 2015
# V1.1.5: Included histo_bin_width(). MC, Oxford, 8 February 2017
# V1.1.6: Added `rasterized` keyword and default value based on points number.
#   Use plt.plot for try values. MC, Oxford, 25 January 2018
# V1.1.7: Changes definition of `extents` for consistency with Scipy's `bounds`.
#   MC, Oxford, 27 April 2018
# V1.1.8: Dropped support for legacy Python 2.7. MC, Oxford, 21 May 2018
#
##############################################################################

__all__ = ["corner_plot", "hist2d", "error_ellipse"]
__version__ = "0.0.6"
__author__ = "Dan Foreman-Mackey (danfm@nyu.edu)"
__copyright__ = "Copyright 2013 Daniel Foreman-Mackey"
__contributors__ = [
    # Alphabetical by first name.
    "Adrian Price-Whelan @adrn",
    "Brendon Brewer @eggplantbren",
    "Ekta Patel @ekta1224",
    "Emily Rice @emilurice",
    "Geoff Ryan @geoffryan",
    "Kyle Barbary @kbarbary",
    "Phil Marshall @drphilmarshall",
    "Pierre Gratier @pirg",
]

import numpy as np
import matplotlib.pyplot as plt
from matplotlib.ticker import MaxNLocator
from matplotlib.colors import LinearSegmentedColormap
from matplotlib.patches import Ellipse
import matplotlib.cm as cm

from plotbin.sauron_colormap  import register_sauron_colormap

######################################################################

def corner_plot(xs, like=None, xstry=None, weights=None, labels=None,
                extents=None, truths=None, truth_color="steelblue", fignum=None,
                scale_hist=False, quantiles=[], verbose=True, init=False,
                plot_contours=True, plot_datapoints=True, fig=None, 
                labels_scaling=1, rasterized=None, **kwargs):
    """
    Make a *sick* corner plot showing the projections of a data set in a
    multi-dimensional space. kwargs are passed to hist2d() or used for
    `matplotlib` styling.

    Parameters
    ----------
    xs : array_like (nsamples, ndim)
        The samples. This should be a 1- or 2-dimensional array. For a 1-D
        array this results in a simple histogram. For a 2-D array, the zeroth
        axis is the list of samples and the next axis are the dimensions of
        the space.
        
    xstry : array_like (nsamples, ndim) (optional)
        Contains all tried parameter, not only the accepted moves.
        
    like : array_like (nsamples) (optional)
        Likelihood of each xs[j, :] set of parameters to be shown on the plot.

    weights : array_like (nsamples,)
        The weight of each sample. If `None` (default), samples are given
        equal weight.

    labels : iterable (ndim,) (optional)
        A list of names for the dimensions.

    extents : 2-tuple of array_like (2, ndim)
        Lower and upper bounds on independent variables. 
        Each array must match the size ndim, e.g.,
        [(0., 10., 30., 25.), (1., 15, 40., 55.)].

    truths : iterable (ndim,) (optional)
        A list of reference values to indicate on the plots.

    truth_color : str (optional)
        A ``matplotlib`` style color for the ``truths`` makers.

    scale_hist : bool (optional)
        Should the 1-D histograms be scaled in such a way that the zero line
        is visible?

    quantiles : iterable (optional)
        A list of fractional quantiles to show on the 1-D histograms as
        vertical dashed lines.

    verbose : bool (optional)
        If true, print the values of the computed quantiles.

    plot_contours : bool (optional)
        Draw contours for dense regions of the plot.

    plot_datapoints : bool (optional)
        Draw the individual data points.

    fig : matplotlib.Figure (optional)
        Overplot onto the provided figure object.

    """
    # Deal with 1D sample lists.
    xs = np.atleast_1d(xs)
    if len(xs.shape) == 1:
        xs = np.atleast_2d(xs)
    else:
        assert len(xs.shape) == 2, "The input sample array must be 1- or 2-D."
        xs = xs.T
        
    if xstry is not None:
        xstry = xstry.T

    if labels is not None:  # note xs was already trasposed
        assert len(labels) == xs.shape[0], 'lengths of labels must match number of dimensions'

    if extents is not None:
        extents = np.array(extents).T
        assert len(extents) == xs.shape[0], 'lengths of extents must match number of dimensions'

    if weights is not None:
        weights = np.asarray(weights)
        assert weights.ndim == 1, 'weights must be 1-D'
        assert xs.shape[1] == weights.shape[0], 'lengths of weights must match number of samples'

    register_sauron_colormap()

    # backwards-compatibility
    plot_contours = kwargs.get("smooth", plot_contours)

    K = len(xs)
    if fig is None:
        factor = 2.0           # size of one side of one panel
        lbdim = 0.5 * factor   # size of left/bottom margin
        trdim = 0.05 * factor  # size of top/right margin
        whspace = 0.05         # w/hspace size
        plotdim = factor * K + factor * (K - 1.) * whspace
        dim = lbdim + plotdim + trdim
        fig, axes = plt.subplots(K, K, figsize=(dim, dim), num=fignum)
        lb = lbdim / dim
        tr = (lbdim + plotdim) / dim
        fig.subplots_adjust(left=lb, bottom=lb, right=tr, top=tr,
                            wspace=whspace, hspace=whspace)
        if init:
            return fig
    else:
        try:
            axes = np.array(fig.axes).reshape((K, K))
        except:
            raise ValueError("Provided figure has {0} axes, but data has "
                             "dimensions K={1}".format(len(fig.axes), K))

    if extents is None:
        extents = [[x.min(), x.max()] for x in xs]

        # Check for parameters that never change.
        m = np.array([e[0] == e[1] for e in extents], dtype=bool)
        if np.any(m):
            raise ValueError(("It looks like the parameter(s) in column(s) "
                              "{0} have no dynamic range. Please provide an "
                              "`extent` argument.")
                             .format(", ".join(map("{0}".format,
                                                   np.arange(len(m))[m]))))
    else:
        # If any of the extents are percentiles, convert them to ranges.
        for i in range(len(extents)):
            try:
                emin, emax = extents[i]
            except TypeError:
                q = [0.5 - 0.5*extents[i], 0.5 + 0.5*extents[i]]
                extents[i] = quantile(xs[i], q, weights=weights)

    # Extract unique values to reduce the memory taken by the figure (MC)
    # The unique version is used for the scatter plots but not for the histograms
    if like is not None:
        likeUnique, w = np.unique(like, return_index=True)
        xsUnique = xs[:, w]

        w = np.argsort(likeUnique) # Sort to plot most likely values last (MC)
        mx = likeUnique[w[-1]]     # Maximum likelihood value so far
        likeUnique = likeUnique[w]
        xsUnique = xsUnique[:, w]

        if rasterized is None:
            rasterized = True if xsUnique.size > 1e4 else False

    for i, x in enumerate(xs):
        ax = axes[i, i]

        # Plot the histograms. Scott's rule for bin size
        # For a Normal distribution sigma=1.4826*MAD, so the robust version of
        # Scott (1979) rule: binsize = 3.49*sigma/n^{1/3} is binsize = 5.17*MAD/n^{1/3}

        # binsize = 5.17*np.median(np.abs(x - np.median(x))) / x.size**(1./3.)
        binsize = 3.49*np.std(x) / x.size**(1./3.)
        # binsize = histo_bin_width(x)
        if binsize > 0:
            nbins = int((extents[i][1] - extents[i][0])/binsize)
            # if nbins > 20:
            #     nbins //= 2  # divide by two for aesthetic reasons and keep integer
            if nbins > 1e4:  # x values are the same within numerical accuracy
                nbins = 5
        else:
            nbins = 5
        ax.cla()  # clean current subplot before plotting
        n, b, p = ax.hist(x, weights=weights, bins=kwargs.get("bins", nbins),
                          range=extents[i], color=kwargs.get("color", "b"),
                          histtype='stepfilled')

        # Axes labels larger than tick labels
        ax.yaxis.label.set_size(plt.rcParams['font.size']*labels_scaling)
        ax.xaxis.label.set_size(plt.rcParams['font.size']*labels_scaling)

        if truths is not None:
            ax.axvline(truths[i], color=truth_color)

        # Plot quantiles if wanted.
        if len(quantiles) > 0:
            qvalues = quantile(x, quantiles, weights=weights)
            for q in qvalues:
                ax.axvline(q, ls="dashed", color=kwargs.get("color", "k"))
            if verbose:
                print("Quantiles:")
                print(zip(quantiles, qvalues))

        # Set up the axes.
        ax.set_xlim(extents[i])
        if scale_hist:
            maxn = np.max(n)
            ax.set_ylim(-0.1 * maxn, 1.1 * maxn)
        else:
            ax.set_ylim(0, 1.1 * np.max(n))
        ax.set_yticklabels([])
        ax.xaxis.set_major_locator(MaxNLocator(5))

        # Not so DRY.
        if i < K - 1:
            ax.set_xticklabels([])
        else:
            [l.set_rotation(45) for l in ax.get_xticklabels()]
            if labels is not None:
                ax.set_xlabel(labels[i])
                ax.xaxis.set_label_coords(0.5, -0.3)

        for j, y in enumerate(xs):
            ax = axes[i, j]

            if j > i:
                ax.set_visible(False)
                ax.set_frame_on(False)
                continue
            elif j == i:
                continue

            ax.cla()  # clean current subplot before plotting

            ax.yaxis.label.set_size(plt.rcParams['font.size']*labels_scaling)
            ax.xaxis.label.set_size(plt.rcParams['font.size']*labels_scaling)

            if xstry is not None:
                ax.plot(xstry[j,:], xstry[i,:], ',m', ms=1, rasterized=rasterized, zorder=0)

            # Plot black when DeltaLogLike=9/2-->DeltaChi2=9 (3sigma)
            if like is not None:
                ax.scatter(xsUnique[j,:], xsUnique[i,:], c=likeUnique,
                           vmin=mx-4.5, vmax=mx, edgecolors='None',
                           cmap='sauron', rasterized=rasterized, **kwargs)
                ax.set_xlim(extents[j])
                ax.set_ylim(extents[i])
            else:
                hist2d(y, x, ax=ax, extent=[extents[j], extents[i]],
                       plot_contours=plot_contours,
                       plot_datapoints=plot_datapoints,
                       weights=weights, **kwargs)

#            ax.tick_params(width=2, which='major')

            if truths is not None:
                ax.plot(truths[j], truths[i], "s", color=truth_color)
                ax.axvline(truths[j], color=truth_color)
                ax.axhline(truths[i], color=truth_color)

            ax.xaxis.set_major_locator(MaxNLocator(5))
            ax.yaxis.set_major_locator(MaxNLocator(5))

            if i < K - 1:
                ax.set_xticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_xticklabels()]
                if labels is not None:
                    ax.set_xlabel(labels[j])
                    ax.xaxis.set_label_coords(0.5, -0.3)

            if j > 0:
                ax.set_yticklabels([])
            else:
                [l.set_rotation(45) for l in ax.get_yticklabels()]
                if labels is not None:
                    ax.set_ylabel(labels[i])
                    ax.yaxis.set_label_coords(-0.3, 0.5)

    return fig

######################################################################

def quantile(x, q, weights=None):
    """
    Like numpy.percentile, but:

    * Values of q are quantiles [0., 1.] rather than percentiles [0., 100.]
    * scalar q not supported (q must be iterable)
    * optional weights on x

    """
    if weights is None:
        return np.percentile(x, [100. * qi for qi in q])
    else:
        idx = np.argsort(x)
        xsorted = x[idx]
        cdf = np.add.accumulate(weights[idx])
        cdf /= cdf[-1]
        return np.interp(q, cdf, xsorted).tolist()

######################################################################

def error_ellipse(mu, cov, ax=None, factor=1.0, **kwargs):
    """
    Plot the error ellipse at a point given its covariance matrix.

    """
    # some sane defaults
    facecolor = kwargs.pop('facecolor', 'none')
    edgecolor = kwargs.pop('edgecolor', 'k')

    x, y = mu
    U, S, V = np.linalg.svd(cov)
    theta = np.degrees(np.arctan2(U[1, 0], U[0, 0]))
    ellipsePlot = Ellipse(xy=[x, y],
                          width=2 * np.sqrt(S[0]) * factor,
                          height=2 * np.sqrt(S[1]) * factor,
                          angle=theta,
                          facecolor=facecolor, edgecolor=edgecolor, **kwargs)

    if ax is None:
        ax = plt.gca()
    ax.add_patch(ellipsePlot)

    return ellipsePlot

######################################################################

def hist2d(x, y, *args, **kwargs):
    """
    Plot a 2-D histogram of samples.

    """
    ax = kwargs.pop("ax", plt.gca())

    extent = kwargs.pop("extent", [[x.min(), x.max()], [y.min(), y.max()]])
    bins = kwargs.pop("bins", 50)
    color = kwargs.pop("color", "k")
    linewidths = kwargs.pop("linewidths", None)
    plot_datapoints = kwargs.get("plot_datapoints", True)
    plot_contours = kwargs.get("plot_contours", True)

    cmap = cm.get_cmap("gray")
    cmap._init()
    cmap._lut[:-3, :-1] = 0.
    cmap._lut[:-3, -1] = np.linspace(1, 0, cmap.N)

    X = np.linspace(extent[0][0], extent[0][1], bins + 1)
    Y = np.linspace(extent[1][0], extent[1][1], bins + 1)
    try:
        H, X, Y = np.histogram2d(x.flatten(), y.flatten(), bins=(X, Y),
                                 weights=kwargs.get('weights', None))
    except ValueError:
        raise ValueError("It looks like at least one of your sample columns "
                         "have no dynamic range. You could try using the "
                         "`extent` argument.")

    # https://corner.readthedocs.io/en/latest/pages/sigmas.html
    # These default levels correspond to [0.5, 1, 1.5, 2]*sigma
    V = 1.0 - np.exp(-0.5 * np.arange(0.5, 2.1, 0.5) ** 2)
    Hflat = H.flatten()
    inds = np.argsort(Hflat)[::-1]
    Hflat = Hflat[inds]
    sm = np.cumsum(Hflat)
    sm /= sm[-1]

    for i, v0 in enumerate(V):
        try:
            V[i] = Hflat[sm <= v0][-1]
        except:
            V[i] = Hflat[0]

    X1, Y1 = 0.5 * (X[1:] + X[:-1]), 0.5 * (Y[1:] + Y[:-1])
    X, Y = X[:-1], Y[:-1]

    if plot_datapoints:
        ax.plot(x, y, "o", color=color, ms=1.5, zorder=-1, alpha=0.1,
                rasterized=True)
        if plot_contours:
            ax.contourf(X1, Y1, H.T, [V[-1], H.max()],
                        cmap=LinearSegmentedColormap.from_list("cmap",
                                                               ([1] * 3,
                                                                [1] * 3),
                        N=2), antialiased=False)

    if plot_contours:
        ax.pcolor(X, Y, H.max() - H.T, cmap=cmap)
        ax.contour(X1, Y1, H.T, V, colors=color, linewidths=linewidths)

    data = np.vstack([x, y])
    mu = np.mean(data, axis=1)
    cov = np.cov(data)
    if kwargs.pop("plot_ellipse", False):
        error_ellipse(mu, cov, ax=ax, edgecolor="r", ls="dashed")

    ax.set_xlim(extent[0])
    ax.set_ylim(extent[1])

######################################################################

def _hist_bin_sturges(x):
    """
    Sturges histogram bin estimator.

    A very simplistic estimator based on the assumption of normality of
    the data. This estimator has poor performance for non-normal data,
    which becomes especially obvious for large data sets. The estimate
    depends only on size of the data.

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.

    Returns
    -------
    h : An estimate of the optimal bin width for the given data.
    """
    return x.ptp() / (np.log2(x.size) + 1.0)

######################################################################

def _hist_bin_fd(x):
    """
    The Freedman-Diaconis histogram bin estimator.

    The Freedman-Diaconis rule uses interquartile range (IQR) to
    estimate binwidth. It is considered a variation of the Scott rule
    with more robustness as the IQR is less affected by outliers than
    the standard deviation. However, the IQR depends on fewer points
    than the standard deviation, so it is less accurate, especially for
    long tailed distributions.

    If the IQR is 0, this function returns 1 for the number of bins.
    Binwidth is inversely proportional to the cube root of data size
    (asymptotically optimal).

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.

    Returns
    -------
    h : An estimate of the optimal bin width for the given data.
    """
    iqr = np.subtract(*np.percentile(x, [75, 25]))
    return 2.0 * iqr * x.size ** (-1.0 / 3.0)

######################################################################

def histo_bin_width(x):
    """

    This function is copied from Numpy 1.11

    Histogram bin estimator that uses the minimum width of the
    Freedman-Diaconis and Sturges estimators.

    The FD estimator is usually the most robust method, but its width
    estimate tends to be too large for small `x`. The Sturges estimator
    is quite good for small (<1000) datasets and is the default in the R
    language. This method gives good off the shelf behaviour.

    Parameters
    ----------
    x : array_like
        Input data that is to be histogrammed, trimmed to range. May not
        be empty.

    Returns
    -------
    h : An estimate of the optimal bin width for the given data.

    See Also
    --------
    _hist_bin_fd, _hist_bin_sturges
    """
    # There is no need to check for zero here. If ptp is, so is IQR and
    # vice versa. Either both are zero or neither one is.
    return min(_hist_bin_fd(x), _hist_bin_sturges(x))

######################################################################



"""
################################################################################

    Copyright (C) 2012-2020, Michele Cappellari
    E-mail: michele.cappellari_at_physics.ox.ac.uk

    Updated versions of the software are available from my web page
    http://purl.org/cappellari/software

    If you have found this software useful for your research,
    I would appreciate an acknowledgement to the use of the
    "adamet Python package of Cappellari et al. (2013), which implements
     the Adaptive Metropolis algorithm of Haario et al. (2001)".

    This software is provided as is without any warranty whatsoever.
    Permission to use, for non-commercial purposes is granted.
    Permission to modify for personal or internal use is granted,
    provided this copyright and disclaimer are included unchanged
    at the beginning of the file. All other rights are reserved.
    In particular, redistribution of the code is not allowed.

################################################################################

MODIFICATION HISTORY:
    V1.0.0: Michele Cappellari, Oxford, 27 January 2012
    V1.1.0: Implemented Adaptive Metropolis Algorithm. MC, Oxford, 10 February 2012
    V1.2.0: Add diagonal matrix to prevent possibly degenerate covariance matrix.
        MC, Oxford, 10 August 2012
    V1.2.1: Do not sort likelihood in output. MC, Oxford, 25 October 2012
    V2.0.0: Translated from IDL into Python . MC, Atlantic Ocean, 9 February 2014
    V2.0.1: Clip input parameters to bounds. MC, Oxford, 18 February 2015
    V2.0.2: Allow for both args and kwargs parameter to the user function.
        Dropped Python 2.7 support. Minor code refactoring.
        MC, Oxford, 13 October 2017
    V2.0.3: First released version. Organized into a package.
        MC, Oxford, 25 April 2018
    V2.0.4: Ensures the user function cannot affect the random chain.
        MC, Oxford, 27 April 2018
    V2.0.5: Some changes to the input format. MC, Oxford, 4 May 2018
    V2.0.6: Fixed clock DeprecationWarning in Python 3.7.
        MC, Oxford, 27 September 2018
    V2.0.7: Written complete docstring documentation.
        MC, Oxford, 4 March 2020

"""

import numpy as np
import matplotlib.pyplot as plt
from time import perf_counter as clock

from adamet.corner_plot import corner_plot

###############################################################################

def _move(pars, sigpars, all_pars, prng):
    """
    Adaptive move

    """
    nu = np.unique(all_pars[:, 0]).size
    npars = pars.size

    # Accumulate at least as many *accepted* moves as the
    # elements of the covariance matrix before computing it.
    if nu > npars*(npars + 1.)/2.:

        eps = 0.01
        diag = np.diag((sigpars*eps)**2)
        cov = 2.38**2/npars*(np.cov(all_pars.T) + diag)
        pars = prng.multivariate_normal(pars, cov)

    else:

        pars = prng.normal(pars, sigpars)

    return pars

###############################################################################

def _metro(lnprob_fun, pars, bounds, lnprob, try_pars, prng, args, kwargs):
    """
    Metropolis step

    """
    # Moves out of bounds are rejected without function evaluation
    if np.all((try_pars >= bounds[0]) & (try_pars <= bounds[1])):

        try_lnprob = lnprob_fun(try_pars, *args, **kwargs)
        lnran = np.log(prng.uniform())
        lnratio = try_lnprob - lnprob

        if (try_lnprob > lnprob) or (lnran < lnratio):

            pars = try_pars   # Candidate point is accepted
            lnprob = try_lnprob

    return pars, lnprob

###############################################################################

def adamet(lnprob_fun, pars, sigpars, bounds, nstep,
           labels=None, nprint=100, quiet=False, fignum=None, plot=True,
           labels_scaling=1, seed=None, args=(), kwargs={}):
    """
    AdaMet Purpose
    --------------

    This is the implementation by
    `Cappellari et al. (2013) <https://ui.adsabs.harvard.edu/abs/2013MNRAS.432.1709C>`_
    of the Adaptive Metropolis algorithm by
    `Haario H., Saksman E., Tamminen J. (2001) <https://doi.org/10.2307/3318737>`_
    for Bayesian analysis.

    Usage Example
    -------------

    To learn how to use ``AdaMet`` run the example program in the
    ``adamet/examples`` directory, within the main package installation
    folder inside ``site-packages``, and read the detailed documentation
    in the docstring of the file ``adamet.py`` or on
    `PyPi <https://pypi.org/project/adamet/>`_.

    Note: For dimensions = 1 to 6, the optimal acceptance rates are
    `rate = [0.441, 0.352, 0.316, 0.279, 0.275, 0.266]`
    and the asymptotic value for many parameters is 23%

    Calling Sequence
    ----------------

    .. code-block:: python

        pars, lnprob = adamet(lnprob_fun, pars0, sigpars0, bounds, nstep,
           labels=None, nprint=100, quiet=False, fignum=None, plot=True,
           labels_scaling=1, seed=None, args=(), kwargs={})

    Input Parameters
    ----------------

    lnprob_fun: callable
        This function returns the natural logarithm of the conditional
        probability of the model, given the data::

            P(model | data) ~ P(data | model) P(model)

    pars0: array_like with shape (n,)
        vector with the mean of the multivariate Gaussian describing the
        proposal distribution from which samples are drawn.
        For maximum efficiency, this initial Gaussian should approximate the
        posterior distribution. This suggests adopting as `pars0` an initial
        guess for the model best-fitting parameters.
    sigpars0: array_like with shape (n,)
        vector with the dispersion `sigma` of the multivariate Gaussian
        describing the proposal distribution.
        For maximum efficiency, this initial Gaussian should approximate the
        posterior distribution. This suggests adopting as `sigpars` an initial
        guess of the uncertainty in the model parameters `pars`.
    bounds: 2-tuple of array_like
        Lower and upper bounds on independent variables. Each array must match
        the size of `pars`. The model probability is set to zero outside the
        bounds. This keyword is also used to define the plotting ranges.
    nsteps: integer
        Number of attempted moves in the chain. Typical numbers are a few
        thousands `nsteps`.


    Optional Keywords
    -----------------

    labels: array_like with shape (n,)
        String labels for each parameter to be used in the `corner_plot`
    nprint: integer
        Specifies the frequency for the intermediate plots, in moves.
        A typical value could be `nstep/10`.
    plot: boolean, optional
        Specifies whether to show a plot of the results or not.
    fignum: integer, optional
        Specifies the figure number for the plot.
    labels_scaling: float
        Relative scaling for the plotting labels.
    seed: integer
        Seed for the random generator. Specify this value for a repeatable
        random sequence.
    args, kwargs: tuple and dict, optional
        Additional arguments passed to `lnprob_fun`. Both empty by default.
        The calling signature is `lnprob_fun(x, *args, **kwargs)`.

    Output Parameters
    -----------------

    pars: array_like with shape (nsteps, n)
        Posterior distribution for the model parameters
    lnprob: array_like with shape (nsteps, n)
        Logarithm of the probbaility of the model, given the data, for each set
        of parameters in the posterior distribution `pars`.

    """

    pars = np.array(pars)  # Make copy to leave input unchanged
    sigpars = np.array(sigpars)
    bounds = np.array(bounds)

    assert pars.size == sigpars.size == bounds.shape[1], \
        "pars, sigpars, and bounds[1] must have the same size"

    if labels is not None:
        assert len(labels) == pars.size, "There must be one label per parameter"

    prng = np.random.RandomState(seed)  # Random stream independent of global one

    t = clock()

    pars = pars.clip(*bounds)  # clip parameters within bounds
    lnprob = lnprob_fun(pars, *args, **kwargs)
    all_pars = np.zeros((nstep, pars.size))
    all_lnprob = np.zeros(nstep)
    all_try_pars = np.zeros_like(all_pars)

    if plot:
        fig = corner_plot(all_pars, init=True, fignum=fignum)
        plt.pause(0.01)

    for j in range(nstep):

        try_pars = _move(pars, sigpars, all_pars[:j], prng)
        pars, lnprob = _metro(lnprob_fun, pars, bounds, lnprob, try_pars, prng, args, kwargs)

        all_pars[j] = pars  # Store only accepted moves or duplicates (when move is rejected)
        all_lnprob[j] = lnprob
        all_try_pars[j] = try_pars  # Store all attempted moves for plotting parameters coverage

        if ((j + 1) % nprint) == 0: # Just plotting/printing inside this block

            nu = np.unique(all_pars[:j, 0]).size
            if quiet is False:
                print('adamet: %0.1f %% done; %0.1f %% accepted' % 
                        (100.*(j + 1)/nstep, 100.*nu/(j + 1)))

            if plot and j > 1:
                corner_plot(all_pars[:j], all_lnprob[:j], xstry=all_try_pars[:j],
                            labels=labels, extents=bounds, fig=fig,
                            labels_scaling=labels_scaling)
                plt.pause(0.01)

    if quiet is False:
        print('adamet: done. Total time %0.2f seconds' % (clock() - t))

    return all_pars, all_lnprob

###############################################################################



from setuptools import setup, find_packages
import re

package = "adamet"

def find_version(package):
    version_file = open(package + "/__init__.py").read()
    rex = r'__version__\s*=\s*"([^"]+)"'
    return re.search(rex, version_file).group(1)

def read_docstring(package):
    main_file = open(package + "/adamet.py").read()
    rex = r'def adamet[\w\W]+?:\n\s*"""([\W\w]+)"""'
    docstring = re.search(rex, main_file).group(1)
    return docstring.replace("\n    ", "\n")

setup(name=package,
      version=find_version(package),
      description="AdaMet: Adaptive Metropolis for Bayesian Analysis",
      long_description=open("README.rst").read()
                       + read_docstring(package)
                       + "\n\nLicense\n-------\n\n"
                       + open("LICENSE.txt").read(),
      long_description_content_type='text/x-rst',
      url="http://purl.org/cappellari/software",
      author="Michele Cappellari",
      author_email="michele.cappellari@physics.ox.ac.uk",
      license="Other/Proprietary License",
      packages=find_packages(),
      package_data={"": ["*.txt"]},
      data_files=["*.txt", "*.rst"],
      install_requires=["numpy", "matplotlib", "plotbin"],
      classifiers=["Development Status :: 5 - Production/Stable",
                   "Intended Audience :: Developers",
                   "Intended Audience :: Science/Research",
                   "Operating System :: OS Independent",
                   "Programming Language :: Python :: 3"],
      zip_safe=True)



#!/usr/bin/env python3

"""
MODIFICATION HISTORY:
    V1.0.0: Michele Cappellari, Oxford, 25 April 2018
    V1.0.1: Adapted for consistency with changes in adamet input.
        MC, Oxford, 4 May 2018

"""

import numpy as np

from adamet.adamet import adamet   

###############################################################################

def lnprob_fun(pars, x, y, sigy):
    """
    The model in this example is a straight line.

    This function returns the natural logarithm of the conditional probability
    of the model, given the data

        P(model | data) ~ P(data | model) P(model)

    In this case we assume:

        1) P(model) = cost (i.e. uninformative prior)
        2) P(data | model) = likelihood for Gaussian errors

    This implies:

        P(model | data) ~ -0.5*chi^2

    """
    ymod = pars[0] + pars[1]*x
    resid = (y - ymod)/sigy
    chi2 = resid @ resid

    # return -0.5*chi2/np.sqrt(2*len(pars))   # ln(likelihood)
    return -0.5*chi2   # ln(likelihood)

###############################################################################

def adamet_example():
    """
    Usage example for the adamet package.
    Fit a straight line.
    
    """
    # Generate some random (x, y) data
    prng = np.random.RandomState(123)  # Give seed for reproducible results
    ndata = 40
    a, b = 10.0, -1.0
    x = prng.uniform(0., 5., ndata)
    sigy = np.ones_like(x)  # Unit errors
    y =  prng.normal(a + b*x, sigy)

    nstep = 10000
    pars0 = np.array([a, b])   # Starting guess

    fargs = (x, y, sigy*(2*len(pars0))**0.25)    # Parameters to pass to the lnprob function
    sigpar = [0.5, 0.5]     # Order of magnitude of the uncertainties
    bounds = np.array([[8, -2], [12, 0]])  # [[min(a), min(b)], [max(a), max(b)]]

    # NB: For the plots to appear properly, one has to to show them in a window,
    # rather than inline, or in a notebook. When running this code from an 
    # IPhython console, one may have to type something like "%matplotlib qt5" 
    # on the console, before running the code.

    # This example matches the optimal acceptance rate of 35% in 2-dim.
    pars, lnprob = adamet(lnprob_fun, pars0, sigpar, bounds, nstep, seed=456,
                        args=fargs, nprint=nstep/10, labels=['a', 'b'])

    bestfit = pars[np.argmax(lnprob)]
    perc = np.percentile(pars, [15.86, 84.14], axis=0)
    sig_bestfit = np.squeeze(np.diff(perc, axis=0)/2)   # half of 68% interval
    print(f"a = {bestfit[0]:0.2f} +/- {sig_bestfit[0]:0.2f}")
    print(f"b = {bestfit[1]:0.2f} +/- {sig_bestfit[1]:0.2f}")

###############################################################################

if __name__ == '__main__':

    adamet_example()

